# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"

import itertools

from TravSHACL.utils.VariableGenerator import VariableGenerator
from TravSHACL.core.RulePattern import RulePattern
from TravSHACL.sparql.QueryGenerator import QueryGenerator


class Shape:
    """This class represents a SHACL shape."""

    def __init__(self, id_, target_def, target_type, target_query, constraints, constraints_id, referenced_shapes,
                 use_selective_queries, max_split_size, order_by_in_queries, include_sparql_prefixes, prefixes=None):
        """
        Creates a new Shape instance representing a SHACL shape that needs to be evaluated.

        :param id_: the name of the shape
        :param target_def: target definition of the shape
        :param target_type: indicates the target type of the shape, e.g., class or node
        :param target_query: target query of the shape
        :param constraints: the constraints belonging to the shape
        :param constraints_id: the constraint ids
        :param referenced_shapes: a list of shapes that is referenced from this shape
        :param use_selective_queries: indicates whether selective queries are used
        :param max_split_size: maximum number of instances per query
        :param order_by_in_queries: indicates whether to use the ORDER BY clause
        :param include_sparql_prefixes: indicates whether to include SPARQL prefixes in queries for the shape
        """
        self.id = id_
        self.constraints = constraints
        self.constraintsId = constraints_id
        self.predicates = []
        self.targetDef = target_def
        self.targetType = target_type          # Might be None
        self.targetQuery = target_query        # Might be None
        self.targetQueryNoPref = target_query  # Might be None
        self.rulePattern = ()
        self.satisfied = None
        self.inDegree = None
        self.outDegree = None

        self.minQuery = None
        self.maxQueries = None
        self.predicates = []
        self.maxValidRefs = {}
        self.skippedQueriesIds = set()

        self.referencedShapes = referenced_shapes
        self.parentShapes = set()
        self.queriesFilters = {}
        self.targets = {"valid": set(), "violated": set()}

        self.useSelectiveQueries = use_selective_queries
        self.querySplitThreshold = max_split_size
        self.ORDERBYinQueries = order_by_in_queries
        self.includePrefixes = include_sparql_prefixes
        if prefixes is None:
            self.prefixes = {}
        else:
            self.prefixes = prefixes
        self.maxConstrId = {}

        self.QueryGenerator = QueryGenerator(self)
        self.__compute_target_queries()

    def get_id(self):
        return self.id

    def set_degree(self, in_, out_):
        self.inDegree = in_
        self.outDegree = out_

    def __compute_predicate_set(self, min_query_id, max_queries_ids):
        """ Returns the ids of the queries for this shape """
        return [self.id] + [self.id + "_d1"] + [min_query_id] + [q for q in max_queries_ids]

    def get_target_query(self):
        return self.targetQuery

    def get_constraints(self):
        return self.constraints

    def get_number_constraints(self):
        """ Gets the number of constraints belonging to this shape """
        return len(self.constraints)

    def get_shape_refs(self):
        return [c.get_shape_ref() for c in self.constraints if c.get_shape_ref() is not None]

    def is_max_zero_ref(self, other_shape_id):
        """ Checks if a referenced shape occurs in a constraint with max zero cardinality """
        ref_constraints = [c for c in self.constraints if c.get_shape_ref() == other_shape_id and c.max == 0]
        return True if ref_constraints else False

    def is_max_ref(self, other_shape_id):
        """ Checks if a referenced shape occurs in a max cardinality constraint """
        ref_constraints = [c for c in self.constraints if c.get_shape_ref() == other_shape_id and c.max != -1]
        return True if ref_constraints else False

    def get_rule_pattern(self):
        return self.rulePattern

    def get_query_split_threshold(self):
        return self.querySplitThreshold

    def __compute_target_queries(self):
        """Internal method to compute the target query of the shape."""
        self.targetQuery = self.QueryGenerator.generate_target_query(
                                        "plain_target",
                                        None,
                                        self.targetQueryNoPref,
                                        self.includePrefixes,
                                        self.ORDERBYinQueries)

        self.queriesFilters = {}
        for ref in self.referencedShapes.keys():
            for c in self.constraints:
                if c.path == self.referencedShapes[ref]:
                    query_valid, query_invalid = self.QueryGenerator.generate_target_query("template_FILTER", [c], self.targetQueryNoPref, self.includePrefixes, self.ORDERBYinQueries)
                    ref_dict = {
                        "query_valid": query_valid,
                        "query_invalid": query_invalid,
                        "constraint": c
                    }
                    self.queriesFilters[ref] = ref_dict

    def compute_constraint_queries(self):
        """Computes all constraint queries for the shape."""
        min_constraints = [c for c in self.constraints if c.min != -1]
        max_constraints = [c for c in self.constraints if c.max != -1]

        subquery = self.QueryGenerator.generate_local_subquery(min_constraints)

        # Build a unique set of triples (+ filter) for all min constraints
        min_id = self.constraintsId + "_pos"
        if len(min_constraints) > 0:
            self.minQuery = self.QueryGenerator.generate_query(
                    min_id,
                    [c for c in min_constraints if c.get_shape_ref() is not None],
                    self.useSelectiveQueries,
                    self.targetQueryNoPref,
                    self.includePrefixes,
                    self.ORDERBYinQueries,
                    subquery
            )

        # Build one set of triples (+ filter) for each max constraint (only one max constraint per query is allowed)
        max_ids = [self.constraintsId + "_max_" + str(i) for i in range(1, len(max_constraints) + 1)]
        for i, max_c in enumerate(max_constraints):
            self.maxConstrId[max_c] = self.constraintsId + "_max_" + str(i+1)

        i = itertools.count()
        self.maxQueries = [self.QueryGenerator.generate_query(
                                        max_ids[next(i)],
                                        [c],
                                        self.useSelectiveQueries,
                                        self.targetQueryNoPref,
                                        self.includePrefixes,
                                        self.ORDERBYinQueries,
                                        subquery) for c in max_constraints]

        self.predicates = self.__compute_predicate_set(min_id, max_ids)
        self.__compute_max_queries_to_skip()
        self.__remove_unevaluated_max_queries()
        self.rulePattern = self.__compute_rule_pattern()

    def __compute_rule_pattern(self):
        """ Computes shape rule pattern """
        focus_node_var = VariableGenerator.get_focus_node_var()
        head = (self.id, focus_node_var, True)

        return RulePattern(head, self.__get_disjunct_rp_body())

    def __get_disjunct_rp_body(self):
        focus_node_var = VariableGenerator.get_focus_node_var()
        min_query = [(self.minQuery.get_id(), focus_node_var, True)] if self.minQuery is not None else []
        max_queries = [(s, focus_node_var, False) for s in [q.get_id() for q in self.maxQueries]]
        return min_query + max_queries

    def add_parent_shape(self, name):
        """ Adds name of incoming neighbor shape in the schema """
        return self.parentShapes.add(name)

    def get_parent_shapes(self):
        return self.parentShapes

    def get_valid_targets(self):
        return self.targets["valid"]

    def get_invalid_targets(self):
        return self.targets["violated"]

    def __compute_max_queries_to_skip(self):
        """
        Creates a dictionary which keys are the names of out-neighbor shapes (defined in max inter-shape constraints)
        of this shape, and the dictionary values are the respective max values of each constraint.
        """
        for max_c in self.get_constraints():
            if max_c.max != -1 and max_c.shapeRef is not None:
                for min_c in self.get_constraints():
                    if min_c.min != -1 and min_c.shapeRef == max_c.shapeRef:
                        self.maxValidRefs[max_c.shapeRef] = max_c.max
                        for constrRef, query_id in self.maxConstrId.items():
                            if constrRef.shapeRef == max_c.shapeRef:
                                self.skippedQueriesIds.add(query_id)

    def __remove_unevaluated_max_queries(self):
        """ Filters max queries to be evaluated against the SPARQL endpoint by excluding needless evaluations """
        self.maxQueries = list(filter(lambda q: q.get_id() not in self.skippedQueriesIds, self.maxQueries))

    def get_max_query_valid_refs(self):
        return self.maxValidRefs

    def get_prefix_string(self):
        return "\n".join(["".join("PREFIX " + key + ": " + value) for (key, value) in self.prefixes.items()]) + "\n"



# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde"

from enum import Enum


class GraphTraversal(Enum):
    """This enum is used to specify the algorithm used for graph traversal."""
    BFS = "Breadth-first search"
    DFS = "Depth-first search"

    def traverse_graph(self, dependencies, reversed_dependencies, starting_point):
        visited = []
        if self == GraphTraversal.DFS:
            self._dfs(visited, dependencies, reversed_dependencies, starting_point)
        elif self == GraphTraversal.BFS:
            self._bfs(visited, dependencies, reversed_dependencies, starting_point)
        return visited

    def _dfs(self, visited, dependencies, reversed_dependencies, node):
        """Implementation of depth-first search with the ability to go back
        if the algorithm is in a sink but there are still unvisited nodes."""
        # TODO: rearrange the graph (do not prioritize dependencies, i.e., use edges = dep + rev_dep)?
        if node not in visited:
            visited.append(node)
            for neighbour in dependencies[node]:
                self._dfs(visited, dependencies, reversed_dependencies, neighbour)
            if sorted(visited) != sorted(dependencies.keys()):
                for neighbour in reversed_dependencies[node]:
                    self._dfs(visited, dependencies, reversed_dependencies, neighbour)
        elif node in visited and sorted(visited) != sorted(dependencies.keys()):
            for neighbour in dependencies[node]:
                if neighbour not in visited:
                    self._dfs(visited, dependencies, reversed_dependencies, neighbour)
            for neighbour in reversed_dependencies[node]:
                if neighbour not in visited:
                    self._dfs(visited, dependencies, reversed_dependencies, neighbour)

    def _bfs(self, visited, dependencies, reversed_dependencies, node):
        """Implementation of breadth-first search.
        Rearranges the graph, i.e., dependencies are not prioritized."""
        edges = self._edges(dependencies, reversed_dependencies)
        queue = [node]
        visited.append(node)
        while queue:
            node = queue.pop(0)
            for neighbour in edges[node]:
                if neighbour not in visited:
                    visited.append(neighbour)
                    queue.append(neighbour)

    @staticmethod
    def _edges(dependencies, reversed_dependencies):
        edges = {}
        for k in dependencies.keys():
            edges[k] = []
            edges[k].extend(dependencies[k])
            edges[k].extend(reversed_dependencies[k])
        return edges



# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde and Monica Figuera"

from TravSHACL.core.ShapeParser import ShapeParser
from TravSHACL.rule_based_validation.Validation import Validation


class ShapeSchema:
    """This class represents a SHACL shape schema."""

    def __init__(self, schema_dir, schema_format, endpoint_url, graph_traversal, heuristics, use_selective_queries,
                 max_split_size, output_dir, order_by_in_queries, save_outputs, work_in_parallel=False):
        """
        Creates a new shape schema instance.

        :param schema_dir: path of the files including the shape definitions
        :param schema_format: indicates the format used for defining the shapes
        :param endpoint_url: URL for the SPARQL endpoint to be evaluated against
        :param graph_traversal: graph traversal algorithm used for determining the evaluation order
        :param heuristics: Python dictionary holding the heuristics used for determining the seed shape
        :param use_selective_queries: indicates whether selective queries are used
        :param max_split_size: maximum number of instances per query
        :param output_dir: output directory for log files
        :param order_by_in_queries: indicates whether to use the ORDER BY clause
        :param save_outputs: indicates whether target classifications will be saved to the output path
        :param work_in_parallel: indicates whether parallelization will be used
        """
        self.shapes = ShapeParser().parse_shapes_from_dir(schema_dir, schema_format, use_selective_queries,
                                                          max_split_size, order_by_in_queries)
        self.shapesDict = {shape.get_id(): shape for shape in self.shapes}  # TODO: use only the dict?
        self.endpointURL = endpoint_url
        self.graphTraversal = graph_traversal
        self.parallel = work_in_parallel  # TODO: no parallelization implemented yet
        self.dependencies, self.reverse_dependencies = self.compute_edges()
        self.compute_in_and_outdegree()
        self.heuristics = heuristics
        self.outputDirName = output_dir
        self.selectivityEnabled = use_selective_queries
        self.saveStats = output_dir is not None
        self.saveTargetsToFile = save_outputs
        self.set_parent_shapes()

    def get_starting_point(self):
        """
        Use heuristics to determine the first shape for evaluation of the constraints.
        There might be several shapes that are equally suited, in this case all of them are returned.

        :return: list of possible starting points based on the heuristics
        """
        possible_starting_points = []

        # heuristic 1: target definition available
        if self.heuristics['target']:
            for s in self.shapes:
                if s.targetDef is not None:
                    possible_starting_points.append(s)

        # heuristic 2: in- and outdegree
        if self.heuristics['degree'] == 'in':
            # prioritize indegree
            possible_starting_points = possible_starting_points if possible_starting_points else self.shapes
            possible_starting_points = self.__indegree(possible_starting_points)
        elif self.heuristics['degree'] == 'out':
            # prioritize outdegree
            possible_starting_points = possible_starting_points if possible_starting_points else self.shapes
            possible_starting_points = self.__outdegree(possible_starting_points)
        elif self.heuristics['degree'] == 'inout':
            # prioritize indegree and further filter by outdegree
            possible_starting_points = possible_starting_points if possible_starting_points else self.shapes
            possible_starting_points = self.__indegree(possible_starting_points)
            possible_starting_points = self.__outdegree(possible_starting_points)
        elif self.heuristics['degree'] == 'outin':
            # prioritize outdegree and further filter by indegree
            possible_starting_points = possible_starting_points if possible_starting_points else self.shapes
            possible_starting_points = self.__outdegree(possible_starting_points)
            possible_starting_points = self.__indegree(possible_starting_points)

        # heuristic 3: number of properties
        if self.heuristics['properties'] == 'small':
            possible_starting_points = possible_starting_points if possible_starting_points else self.shapes
            if len(possible_starting_points) > 1:
                min_con = min([s.get_number_constraints() for s in possible_starting_points])
                tmp = []
                for s in possible_starting_points:
                    if s.get_number_constraints() == min_con:
                        tmp.append(s)
                possible_starting_points = tmp
        elif self.heuristics['properties'] == 'big':
            if len(possible_starting_points) > 1:
                max_con = max([s.get_number_constraints() for s in possible_starting_points])
                tmp = []
                for s in possible_starting_points:
                    if s.get_number_constraints() == max_con:
                        tmp.append(s)
                possible_starting_points = tmp

        if not possible_starting_points:
            possible_starting_points = all
        return [s.get_id() for s in possible_starting_points]

    @staticmethod
    def __indegree(possible_starting_points):
        if len(possible_starting_points) > 1:
            max_in = max([s.inDegree for s in possible_starting_points])
            tmp = []
            for s in possible_starting_points:
                if s.inDegree == max_in:
                    tmp.append(s)
            return tmp
        else:
            return possible_starting_points

    @staticmethod
    def __outdegree(possible_starting_points):
        if len(possible_starting_points) > 1:
            max_out = max([s.outDegree for s in possible_starting_points])
            tmp = []
            for s in possible_starting_points:
                if s.outDegree == max_out:
                    tmp.append(s)
            return tmp
        else:
            return possible_starting_points

    def set_parent_shapes(self):
        """Sets the shape referring to a shape as its parent."""
        for shape_name in self.shapesDict:
            child_shapes = self.shapesDict[shape_name].referencedShapes
            for child_name in child_shapes:
                self.shapesDict[child_name].add_parent_shape(shape_name)

    def validate(self):
        """Executes the validation of the shape network."""
        start = self.get_starting_point()
        node_order = self.graphTraversal.traverse_graph(self.dependencies, self.reverse_dependencies, start[0])  # TODO: deal with more than one possible starting point

        for s in self.shapes:
            s.compute_constraint_queries()

        target_shapes = [s for name, s in self.shapesDict.items()
                         if self.shapesDict[name].get_target_query() is not None]
        target_shape_predicates = [s.get_id() for s in target_shapes]

        return Validation(
            self.endpointURL,
            node_order,
            self.shapesDict,
            target_shape_predicates,
            self.selectivityEnabled,
            self.outputDirName,
            self.saveStats,
            self.saveTargetsToFile
        ).exec()
        # return 'Go to log files in {} folder to see report'.format(self.outputDirName)

    def compute_in_and_outdegree(self):
        """Computes the in- and outdegree of each shape."""
        for s in self.shapes:
            s.set_degree(
                in_=len(self.reverse_dependencies[s.get_id()]) if s.get_id() in self.reverse_dependencies.keys() else 0,
                out_=len(self.dependencies[s.get_id()]) if s.get_id() in self.dependencies.keys() else 0
            )
        return

    def compute_edges(self):
        """Computes the edges in the network."""
        dependencies = {s.get_id(): [] for s in self.shapes}
        reverse_dependencies = {s.get_id(): [] for s in self.shapes}
        for s in self.shapes:
            refs = s.get_shape_refs()
            if refs:
                name = s.get_id()
                dependencies[name] = refs
                for ref in refs:
                    reverse_dependencies[ref].append(name)
        return dependencies, reverse_dependencies



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera"


class RulePattern:
    """Implementation of Datalog-like rules."""

    def __init__(self, head, body):
        self.__head = head
        self.__literals = body
        self.__variables = set([head[1]] + [a[1] for a in body if a is not None])

    def __repr__(self):
        return "[" + str(self.head) + "] <-- " + str(self.body)

    @property
    def head(self):
        return self.__head

    @property
    def body(self):
        return self.__literals

    @property
    def variables(self):
        return self.__variables



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"

from TravSHACL.utils.VariableGenerator import VariableType
from TravSHACL.constraints.Constraint import Constraint


class MaxOnlyConstraint(Constraint):
    """This class represents max constraints, i.e., a constraint for the maximal occurrence of a path."""

    def __init__(self, var_generator, id_, path, max_, is_pos, datatype=None, value=None, shape_ref=None, target_def=None):
        """
        Creates a new max constraint.

        :param var_generator: variable generator instance
        :param id_: name of the constraint
        :param path: the path associated with this constraint, e.g., a predicate
        :param max_: the maximal occurrence allowed
        :param is_pos: true if it is a positive constraint, false otherwise
        :param datatype: contains the datatype the object must fulfill
        :param value: contains the value the constraint checks again, i.e., an object
        :param shape_ref: contains the name of the shape referenced by the constraint, none otherwise
        :param target_def: contains the target definition of the shape the constraint belongs to if it has one
        """
        super().__init__(id_, is_pos, None, datatype, value, shape_ref, target_def, path)
        self.varGenerator = var_generator
        self.min = -1
        self.max = max_
        self.variables = self.compute_variables()

    def compute_variables(self):
        """Computes variable names for the SPARQL queries of the constraint."""
        atomic_constraint = Constraint()
        return atomic_constraint.generate_variables(self.varGenerator, VariableType.VALIDATION, self.max + 1)



# -*- coding: utf-8 -*-

import itertools
from enum import Enum

i = itertools.count()


class VariableGenerator:
    """Used to create unused variables for the SPARQL queries."""

    def __init__(self):
        self.index = 0

    @staticmethod
    def generate_variable(type_):
        type_ = "p_"  # *** hardcoded
        return str(type_) + str(next(i))

    @staticmethod
    def get_focus_node_var():
        return "x"  # *** hardcoded


class VariableType(Enum):
    VALIDATION = "p_"
    VIOLATION = "n_"  # not used

    @classmethod
    def prefix(cls):
        return cls.VALIDATION



# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde"

from SPARQLWrapper import SPARQLWrapper, JSON


class SPARQLEndpoint:
    """Implementation of a SPARQL endpoint; having a URL and the ability to run SPARQL queries."""

    class __SPARQLEndpoint:
        """Private class to allow simulation of a Singleton."""
        def __init__(self, endpoint_url):
            self.endpointURL = endpoint_url
            self.endpoint = SPARQLWrapper(endpoint_url)
            self.endpoint.setReturnFormat(JSON)

        def run_query(self, query_string):
            self.endpoint.setQuery(query_string)
            return self.endpoint.query().convert()

    instance = None

    def __new__(cls, endpoint_url):
        if not SPARQLEndpoint.instance:
            SPARQLEndpoint.instance = SPARQLEndpoint.__SPARQLEndpoint(endpoint_url)
        return SPARQLEndpoint.instance

    def __getattr__(self, item):
        return getattr(self.instance, item)

    def __setattr__(self, key, value):
        return setattr(self.instance, key, value)



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"


class Constraint:
    """Base class for all constraints."""

    def __init__(self, id_=None, is_pos=None, satisfied=None, datatype=None, value=None,
                 shape_ref=None, target_def=None, path=None):
        """
        Base constructor for all constraints.

        :param id_: name of the constraint
        :param is_pos: true if it is a positive constraint, false otherwise
        :param satisfied: indicates whether the constraint is satisfied, should be unknown at creation
        :param datatype: contains the datatype the object must fulfill
        :param value: contains the value the constraint checks again, i.e., an object
        :param shape_ref: contains the name of the shape referenced by the constraint, none otherwise
        :param target_def: contains the target definition of the shape the constraint belongs to if it has one
        :param path: the path associated with this constraint, e.g., a predicate
        """
        self.id = id_
        self.isPos = is_pos
        self.satisfied = satisfied

        self.datatype = datatype
        self.value = value
        self.shapeRef = shape_ref
        self.target = target_def

        self.variables = []
        self.path = path

    def get_datatype(self):
        return self.datatype

    def get_value(self):
        return self.value

    def get_shape_ref(self):
        return self.shapeRef

    def get_id(self):
        return self.id

    def get_is_pos(self):
        return self.isPos

    @staticmethod
    def generate_variables(var_generator, type_, number_of_variables):
        """Generates variable names for the SPARQL queries of the constraint."""
        vars_ = []
        for elem in range(number_of_variables):
            vars_.append(var_generator.generate_variable(type_))

        return vars_

    def get_variables(self):
        return self.variables

    def compute_rule_pattern_body(self):
        """
        Compute the body of the rule patterns representing the constraint.

        :return: rule pattern body of the constraint
        """
        return [(self.shapeRef, v, self.isPos) for v in self.variables] if self.shapeRef is not None else []
